X : AbstractFunction {

	var <source, <>children;
	var <>updateFunction;
	
	*new { arg obj, updateFunction;
		^super.new.addToFamily(obj).updateFunction_(updateFunction)
	}
	
	*wrapObj { arg obj;
		 ^if(obj.isKindOf(X).not) {X.newCopyArgs(obj) } {obj };
	}
	
	source_ { arg obj;
		this.addToFamily(obj);
		this.update(this);
	}
	dereference {arg alreadyChecked;
		alreadyChecked = alreadyChecked ?? { IdentitySet.new };
		if(alreadyChecked.includes(this)) { "cyclic graph cannot be dereferenced".warn; ^this };
		^source.dereference(alreadyChecked);
	}
	
	addToFamily { arg obj;
		obj = this.class.wrapObj(obj);
		if(source.notNil) { source.removeFrom(this) };
		source = obj;
		obj.children = obj.children.add(this);
	}
	removeFrom { arg from;
		if(children.notNil) { children.remove(from) };
	}
	
	
	update { arg byWhom, alreadyChecked;
		alreadyChecked = alreadyChecked ?? { IdentitySet.new };
		if(alreadyChecked.includes(this).not) {
			updateFunction.value(byWhom);
			alreadyChecked.add(this);
			children.do { arg item; item.update(this, alreadyChecked) };
		};
		
	}
	
	// function composition
	
	value { ^this.dereference.value }
	
	composeUnaryOp { arg aSelector;
		^UnaryOpX.new(aSelector, this)
	}
	composeBinaryOp { arg aSelector, something;
		^BinaryOpX.new(aSelector, this, something)
	}
	reverseComposeBinaryOp { arg aSelector, something;
		^BinaryOpX.new(aSelector, something, this)
	}
	composeNAryOp { arg aSelector, anArgList;
		^thisMethod.notYetImplemented
	}
	
	
	
	printOn { arg stream; stream << this.class.name << "(" << source << ")" }
	
	// usages
	
	play {
		var synth;
		^X.new(this, { arg x; 
			synth.release;
			synth = x.dereference.play 
		});
	}
	
	playInNodeProxy { arg proxy;
		^X.new(this, { arg x;
			proxy.source = x.dereference;
		})
	}
	
	watcher {
		^X.new(this, { arg x;
			[this, x.dereference].postln;			
		})
	}

}

AbstractOpX : X {
	composeUnaryOp { arg aSelector;
		^UnaryOpX.new(aSelector, this)
	}
	composeBinaryOp { arg aSelector, something;
		^BinaryOpX.new(aSelector, this, something)
	}
	reverseComposeBinaryOp { arg aSelector, something;
		^BinaryOpX.new(aSelector, something, this)
	}
	composeNAryOp { arg aSelector, anArgList;
		^{ this.value(anArgList).performList(aSelector, anArgList) }
	}
	
}


UnaryOpX : AbstractOpX {
	var >operator;
	
	*new { arg operator, a;
		^super.new(a).operator_(operator)
	}
	dereference { arg alreadyChecked; 
		alreadyChecked = alreadyChecked ?? { IdentitySet.new };
		if(alreadyChecked.includes(this)) { "cyclic graph cannot be dereferenced".warn; ^this };
		alreadyChecked.add(this);
		^source.dereference(alreadyChecked).perform(operator) 
	}

}

BinaryOpX : AbstractOpX {
	var >operator, >argument;
	*new { arg operator, a, b;
		^super.new([a,b]).operator_(operator).argument_(b)
	}
	addToFamily { arg ab;
		var a, b;
		#a, b = ab;
		a = this.class.wrapObj(a);
		b = this.class.wrapObj(b);
		source = a;
		argument = b;
		a.children = a.children.add(this);
		b.children = b.children.add(this);
	}
	dereference { arg alreadyChecked;
		var vala, valb;
		alreadyChecked = alreadyChecked ?? { IdentitySet.new };
		if(alreadyChecked.includes(this)) { "cyclic graph cannot be dereferenced".warn; ^this };
		alreadyChecked.add(this);
		vala = source.dereference(alreadyChecked);
		valb = argument.dereference(alreadyChecked);
		^vala.perform(operator, valb) 
	}
	
}


